use nix::unistd::ftruncate;
use std::fs::OpenOptions;
use std::io::{Read, Seek, SeekFrom, Write};
use std::thread;
use std::time::Instant;
use std::{cmp, fs};
use std::{env, fs::File, os::fd::IntoRawFd};

const NUM_THREADS: i64 = 8;

fn truncate_dst(f: File, len: i64) -> nix::Result<()> {
    ftruncate(f.into_raw_fd(), len)
}

fn copy_range(src: &mut File, dst: &mut File, offset: u64, len: u64, chunk_bytes: u64) {
    src.seek(SeekFrom::Start(offset)).unwrap();
    dst.seek(SeekFrom::Start(offset)).unwrap();

    let end = cmp::min(offset + chunk_bytes, len);

    // replace chunk_bytes with the actual nr of chunk_bytes
    // which can be less if we are at the last chunk of src
    let chunk_bytes = (end - offset) as usize;

    let bufsz = cmp::min(1024 * 1024, chunk_bytes);
    let mut buf = vec![0u8; bufsz];

    let nr_of_full_reads = chunk_bytes / bufsz;
    let bytes_for_last_read = chunk_bytes % bufsz;

    // full reads
    for _ in 0..nr_of_full_reads {
        src.read_exact(&mut buf).unwrap();
        dst.write_all(&buf).unwrap();
    }

    // last (partial) read
    if bytes_for_last_read > 0 {
        buf.resize(bytes_for_last_read, 0);
        src.read_exact(&mut buf).unwrap();
        dst.write_all(&buf).unwrap();
    }
}

fn thread_copy_chunks(thread_nr: i64, src: &str, dst: &str, len: i64) {
    let mut src = File::open(src).unwrap();
    let mut dst = OpenOptions::new().write(true).open(dst).unwrap();

    let chunk_bytes = len / (NUM_THREADS * 2);

    let jump = NUM_THREADS * chunk_bytes;

    let mut offset = thread_nr * chunk_bytes;
    let mut loop_count = 0;
    loop {
        println!("t{thread_nr} l{loop_count} {offset} chunk={chunk_bytes}");

        copy_range(
            &mut src,
            &mut dst,
            offset as u64,
            len as u64,
            chunk_bytes as u64,
        );

        offset += jump;
        if offset >= len {
            break;
        }

        loop_count += 1;
    }
}

fn main() {
    let src = env::args().nth(1).unwrap();
    let dst = env::args().nth(2).unwrap();

    let len = fs::metadata(&src).unwrap().len();
    if len > i64::MAX as u64 {
        panic!("src len > i64::MAX");
    }

    {
        let dst = File::create(&dst).unwrap();
        truncate_dst(dst, len as i64).unwrap();
    }

    println!("Spawning {NUM_THREADS} threads to copy {src} to {dst} ({len} bytes)");

    let now = Instant::now();
    // create threads in scope so borrows from the main thread
    // outlive all threads
    thread::scope(|s| {
        for thread_nr in 0..NUM_THREADS {
            let src_ref = &src;
            let dst_ref = &dst;
            s.spawn(move || thread_copy_chunks(thread_nr, src_ref, dst_ref, len as i64));
        }
    });

    println!("took {:.3}s", now.elapsed().as_secs_f64());
}
